/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/

// Type declarations global to everything

#ifndef GLOBAL_H
#define GLOBAL_H

#include <inttypes.h>
#include <avr/pgmspace.h>

/* special hardware is defined in my_make (see doc/)
*/
#ifdef GENBOARDv3
#define HWMAP
#include "hwdef_v3.h"
#else
#include "hwdef_v2.h"
#endif

#define INJ_PER_REV 2

#ifndef true 
#define true 1
#endif

#ifndef false
#define false 0
#endif


/***********************************************************************/
//flags
// #define tenth_second 0
#define second       1

// the high resolution now and now_upper (16+16 bits) is in eventqueue
struct time_t {
// these are clocked from handler2k:
//  uint8_t iac_clk;
  uint8_t adc_clk;	
//  uint8_t tsec_clk; // 100 ms clock
//  uint16_t h2k;   // h2khandler counter (800 usec) is in benchstats. Move here if you need it.
// these are clocked from userspace:
  uint8_t tsec;  // seconds /10
  uint16_t sec;  // seconds
// written from both userspace(use cli-sei!) and interrupt.
  uint8_t flags; 
};

// rtc is not volatile any longer. Some fields are only set and beeing watched
// from userspace (like flags and most); adc_clk only from handler2k()
extern struct time_t rtc;
/***********************************************************************/

#define enabled   0 // squirt
#define scheduled 1 // sched
#define firing    2 // firing

struct squirt_t {
	uint16_t pwcalc;    // pulsewidth in 4 usec resolution
#ifndef GENBOARDv3
	uint8_t pwm_dc;    // pwm duty cycle (injpwm)
#endif
	uint16_t pwm_delay; // delay (4 usec) before going to pwm-mode (injpwmt) SHOULD depend on Vbatt
	uint16_t stopat;	 // when to stop (4 usec time)
};

// volatile dropped, interrupt will reload the values anyway
#ifdef GENBOARDv3
#define INJECTORS 8
#else
#define INJECTORS 2
#endif

extern struct squirt_t inj_port[INJECTORS];

/***********************************************************************/

//status
#define running 0  // engine running
#define crank   1  // cranking
#define startw  2  // start warmup enrichment
#define warmup  3  // warmup
#define tpsaen  4  // tps acceleration
#define tpsden  5  // tps deceleration
//#define mapaen  6  // map acceleration
#define idle    7  // idle mode

//status_ext
#define new_rpm          0  // need for recalculation of rpm
#define baro_problem     1  // somehow the barometric pressure is way off
#define o2_not_ready     2  // this is set for the first 30 seconds
#define o2_closed_loop   3  // enabled when o2 sensor is active
#define crank_enable     4  // allowed to enter crank mode
#define wheel_in_sync    5  // tooth wheel in sync
#define wheel_error      6  // error during wheel decoding
#define early_dwellstart 7	// dwellstart_before_trigger

struct engine_t {
  uint8_t status;      // uses the defines above
  uint8_t status_ext;  // uses the defines below the above!
  uint8_t kpa;         // MAP value in units of KPa
  uint8_t coolant;     // Coolant temperature in Degrees F plus 40 (allows -40 degress to fit in integer)
  uint8_t mat;         // manifold air temperature
  uint8_t batt;        // current battery voltage
  uint8_t tps;         // throttle posision (percent)
  uint8_t last_tps;    // throttle posision (percent), last reading
  uint32_t rpm_period; // rpm period (in 4 usec resolution)
  uint8_t rpm;         // computed engine rpm/100 (stupid origins in motorola megasquirt, comm compatibility likes this)
//  uint16_t rpm_64;     // computed highres engine rpm: 64 * rpm/100 obsolete, see rpm_vhr
  uint16_t rpm_vhr;		// computed very high res RPM (real RPM, like 8000)
  uint8_t current_ignadv;  // measured ignition advance (toothwheel + coil trigger)
  uint8_t current_dwell;  // measured ignition dwell
#ifndef NOIGN
  uint8_t ignadv;			// ignition advance from table interpolation
  uint8_t ignstate;		// see values in timing.c
  uint16_t dwell;		// in 4 usec resolution. (This could be 64 usec resolution and uint8_t?)
  uint16_t trig2spark;	// time from trigger to spark = rpm_period * trig2spark / 4096
  uint16_t sparkat;	// trigger fills it in knowing fresh period value
#endif
};

// max value of engine.rpm_trigger_reenable
#define TRIGGER_REENABLE_MAX 5000   // 5000*4us = 20.0 ms

extern volatile struct engine_t engine;
/***********************************************************************/

// define the adc sampling period [divisor from 1250Hz]
// THIS SHOULD MATCH H2KDIVIDE in eventqueue.h (maybe make it automatic?)
#define ADC_PERIOD  6

// TIMER3 TOP=1023. TIMER1 could use the same TOP value,
// but a little different value might be nice on the supply.
// The same frequency and artificial offset between TCNT3 and TCNT1 might be the best?
// PWM_FREQUENCY = 16MHz / 15.984kHz - 1
#define PWM_FREQUENCY   1000

// n-1 seconds after leaving crankmode we can't reenter (unless rpm drops to 0)
#define CRANK_TIMEOUT 11

// tps threshold
#define FLOOD_CLEAR 155

// version number of B&G MegaSquirt
#define BG_EMULATION 20

// these should teoretically make it harder to make stupid bugs!
#define COLD_TEMP 0     // -40 F == -40 C
#define WARM_TEMP 205   // 165 F == 74 C (assume the engine is warm at this temperature)

#define BATT_7V 61
#define BATT_14V 118
/***********************************************************************/
/* Configuration variables */
/***********************************************************************/

// DO NOT PUT IFDEFS here unless you teach stru.pl to parse my_make
// Otherwise magicstr will be offsetted, and you will have a bad day. You have been warned.
struct config_t {

  uint8_t primep;               // Priming pulse at -40 F [100 usec]
  uint8_t primep_temp_scaling;  // Priming pulse scaling at 170 F (80=half, ff=same as cold)

  uint8_t cwl;            // Cranking pulsewidth [100 usec] at -40 F
  uint8_t cwh;            // Cranking pulsewidth [100 usec] at 170 F
  uint8_t cranking_thres; // Cranking threshold [100 rpm]

  uint8_t awev;         // Afterstart warmup enrichment (percent, added value)
  uint8_t awev_temp_scaling; // Afterstart warmup scaling at 170F (80=half, ff=same as cold)
  uint8_t awc;          // Afterstart number of engine cycles

  uint8_t warmup_clt_range[10];  // Warmup temperature range of bins (+40 F offset)
  uint8_t warmup_clt[10];        // Warmup enrichment bins(fn temp), biased at 100
  uint8_t warmup_rpm[8];         // Warmup enrichment bins(fn rpm), biased at 100

  uint8_t req_fuel;     // Fuel constant (req_fuel)
  uint8_t divider;      // IRQ divide factor for pulse
  uint8_t alternate;    // Alternate injector drivers
  uint8_t injopen;      // Injector open time at 13.2V [16 usec based]
  uint8_t battfac;      // Battery gamma factor [16 usec based]
  uint8_t injocfuel;    // PW-correlated amount of fuel injected during injector open [16 usec based]
  uint8_t injpwm;       // Injector PWM duty cycle (when current limiting is active)
  uint8_t injpwmt;      // Injector PWM time at which to activate pwm [100 usec]
  uint8_t injpwm6;	// added Injector PWM duty cycle during low battery voltage (6V)

  uint8_t rpmk[2];      // Constant for calculating, rpm = 12,000/ncyl (0: high byte)

  uint8_t tpsdotrate[4];  // TPSDOT acceleration bins, dv/dt
  uint8_t tpsaq[4];       // TPS acceleration amount (fn TPSDOT) [100 usec]

  uint8_t tps_thresh;   // Accel TPSDOT threshold (throttle sensitivity)
  uint8_t tpsasync;     // Acceleration duration [0.1 sec]
  uint8_t acmult;       // Acceleration cold multiplication factor, biased at 100
  uint8_t tpsacold;     // Cold acceleration added amount (at -40 F degrees) [100 usec]
  uint8_t tpsdq;        // Deacceleration fuel cut, biased at 100
  uint8_t decel_fuelcut_thres;  // Decel fuelcut enabled above threshold [100 rpm]

  uint8_t overrun_fuelcut;        // Overrun fuelcut, injectors disabled above this [100 rpm]
  uint8_t overrun_fuelresume;     // Overrun fuelresume, injectors reenabled below this [100 rpm]

  uint8_t rev_limit;      // Rev limit (hard) [100 rpm]

  uint8_t airden_ignore;  // Fake airdensity low limit (low rpm only), 0==disable

  uint8_t config11;     // c11, bit0: 0:mpx4115/1:mpx4250 bit1: ...
  uint8_t config12;     // c12, bit0: ...
  uint8_t config13;     // c13, bit0: ...

  uint8_t mt_unused[4]; // Unused, bonus bytes used by megatune

  uint8_t fastidle;     // Fastidle valve off above this Temp[..F-40]

  uint8_t baro;           // mean barometric pressure
  uint8_t dbaro;          // max difference in barometric pressure

  uint8_t tps_low;        // tps, min adc-count
  uint8_t tps_high;       // tps, max adc-count

  uint8_t fan_temp;       // coolant fan, start temperature
  uint8_t fan_hyst;       // coolant fan, hysteresis

  // IAC stuff
  uint8_t iac_step_seq;        // iac, stepping sequence (define iac wiring)
  uint8_t iac_conf;            // iac, configuration (stepper speed, power-off)
  uint8_t iac_max_steps;       // iac, max steps to extend
  uint8_t iac_tps_thres;       // iac, engine.tps value for idle threshold
  uint8_t iac_cold_idle_temp;  // iac, max temperature for cold idle speed
  uint8_t iac_warm_idle_temp;  // iac, min temperature for warm idle speed
  uint8_t iac_cold_rpm;        // iac, idle rpm (cold) [10 rpm]
  uint8_t iac_warm_rpm;        // iac, idle rpm (warm) [10 rpm]
  uint8_t iac_cold_start_pos;  // iac, crank position at -40 F
  uint8_t iac_warm_start_pos;  // iac, crank position at 170 F
  uint8_t iac_afterstart_rpm;  // iac, afterstart rpm [10 rpm], added to idle rpm
  uint8_t iac_afterstart_duration; // iac, duration of afterstart rpm [0.1 sec]
  uint8_t iac_afterstart_steps; // iac, afterstarts: steps per 160 rpm
  uint8_t iac_kp;              // iac, P-term
  uint8_t iac_ki;              // iac, I-term
  uint8_t iac_kd;              // iac, D-term
  uint8_t iac_integral_speed;  // iac, integral time speed
  uint8_t iac_integral_limit_dec;  // iac, integral decrease limit
  uint8_t iac_integral_limit_inc;  // iac, integral increase limit
  uint8_t iac_integral_deadband; // iac, integral deadband [1 rpm] (being tested for usefulness)
  uint8_t iac_deadband;        // iac, deadband [1 rpm]
  uint8_t iac_pid_conf;        // iac, config, 0:asymmetric
  uint8_t iac_overclose_interval;  // iac, seconds between overclosing
  uint8_t iac_ref_pos[10];     // iac, reference position (fn temp)

  // ego controller
  uint8_t ego_conf;        // ego, config bit0:PID bit1:WBO2 bit2:incredible bit4:NBO2-l-target
  uint8_t ego_delta;       // ego, step size (percent) [0.4%]
  uint8_t ego_lag;         // ego, timeconstant (engine cycles) for changing correction factor
  uint8_t ego_coolant;     // ego, min coolant temperature for ego enabling[F]
  uint8_t ego_maxtps;      // ego, max tps position for ego correction
  uint8_t ego_maxmap;      // ego, max map for O2 correction[kPa]
  uint8_t ego_minrpm;      // ego, min rpm for ego enabling [100rpm]
  uint8_t ego_maxrpm;      // ego, max rpm for O2 correction [100rpm]
  uint8_t ego_warmup;      // ego, warmup time [sec]
  uint8_t ego_lean_limit;  // ego, max percent to lean afr[0.4%](in SINGLE_TARGET mode rich too)
  uint8_t ego_rich_limit;  // ego, max percent to rich afr[0.4%]
  uint8_t ego_target;      // ego, target voltage (NBO2)
  uint8_t ego_pid_kp;      // ego, pid kp
  uint8_t ego_pid_ki;      // ego, pid ki
  uint8_t ego_pid_kd;      // ego, pid kd
  uint8_t ego_pid_ilimit;  // ego, pid integral limit
  uint8_t ego_pid_window;  // ego, minimum window for pid to react

  // WBO2
  // heater:
  uint8_t wbo2_warmup_ramp;     // wbo2, warmup ramp speed [3f=0.2V/sec]
  uint8_t wbo2_warmup_target;   // wbo2, warmup target
  uint8_t wbo2_abs_limit;	// wbo2, heater absolute limit [2a=12V ff=30V]
  uint8_t wbo2_limit_maxt;     // wbo2, max time spent at abs_limit [262msec]
  uint8_t wbo2_fallback;     // wbo2, fallback V after maxt [1c=10V ff=30V]
  uint8_t wbo2_retry_t;     // wbo2, retry control after [4sec]

  // measurement and controller
  uint8_t wbo2_edgetime_corr;	// wbo2, edgetime correction constant C/R 
  uint8_t wbo2_edgetime_min;		// wbo2, smaller edgetimes are neglected [0.5usec]
  uint8_t wbo2_ri_target;	// wbo2, target Ri (pulse amplitude:C3)[5V/4096]
  uint8_t wbo2_nernstdc_target; // wbo2, target nernstDC [5V/256]
  uint8_t wbo2_pump_pw_zero;    // wbo2, pump zero dutycycle (default: 66) [5V/256]
  uint8_t wbo2_calibration;     // wbo2, calibration constant
  uint8_t wbo2_heater_pid_kp;   // wbo2, heater pid kp
  uint8_t wbo2_heater_pid_ki;   // wbo2, heater pid ki
  uint8_t wbo2_heater_pid_kd;   // wbo2, heater pid kd
  uint8_t wbo2_heater_pid_ilimit;   // wbo2, heater pid integral limit

  uint8_t wbo2_pump_pid_kp;   // wbo2, pump pid kp
  uint8_t wbo2_pump_pid_ki;   // wbo2, pump pid ki
  uint8_t wbo2_pump_pid_kd;   // wbo2, pump pid kd
  uint8_t wbo2_pump_pid_ilimit;   // wbo2, pump pid integral limit

  uint8_t wbo2_ri_confidence_scale;  // wbo2, Ri confidence scaling factor (s^2 < 255/scale)


  // VE learning stuff
  uint8_t ve_learn_coolant;      // VE learn, min coolant for auto modification of VE entries
  uint8_t ve_learn_max_power;    // VE learn, auto modification of VE entry only below power=map*rpm/256
  uint8_t ve_learn_rpm_scale;    // VE learn, rpm scaling factor (s^2 < 255/scale)
  uint8_t ve_learn_kpa_scale;    // VE learn, kpa scaling factor (s^2 < 255/scale)
  uint8_t ve_learn_ego_scale;    // VE learn, ego scaling factor (s^2 < 255/scale)
  uint8_t ve_learn_speed;        // VE learn, speed of changes to the VE table
  uint8_t ve_learn_limit;        // VE learn, max deviation (+/- [0.4%]) from reference VE table
  uint8_t ve_learn_conf;         // VE learn, config, bit0:enable bit1:simulate bit2: 2x2box

  uint8_t lcd_c0;			// LCD flags 2:phasesignoff 1:lcd_easygoto 0:busypolled 
  uint8_t lcd_delay;			// LCD-cablelength delays Higherhalf:beforeENable LowerHalf:ENable
  uint8_t lcd_backlight;		// LCD backlight: FF is fully on
  uint8_t lcd_offs[4];			// line addresses for LCD.ff=default
  uint8_t lcd_default_view;             // LCD, default runtime view
  
  uint8_t primary_trigger;      // trigger1, bit0 0:falling/1:rising, bit1 0:toothwheel/1:coil, bit2 0:no filtering/1:filtering
  uint8_t secondary_trigger;    // trigger2, bit0 0:falling/1:rising, bit1 0:enable/1:disable, bit2 0:no filtering/1:filtering

  uint8_t tooth_wheel;          // tooth wheel, number of teeth on the wheel
  uint8_t trigger_tooth;        // tooth wheel, active trigger tooth (00:missing tooth)
  uint8_t crank_minper;		// Cranktigger period minimum (*16 usec)

  uint8_t ign_tdcdelay;		// TDC after the trigger(0.5crankdeg)
  uint8_t ign_dwell14;		// dwell above 14V (64usec)
  uint8_t ign_dwell6;		// added dwell time at 6V (27usec)
  uint8_t ign_crank_advance;    // ignition advance at cranking [0.25crankdeg]
  // developer bit-definition of ign_out:
  // 7:disable, 6:EDIS/dummy, 5:after/before, 4:ref_to_trigger/ref_to_spark 3:multispark
  // EDIS bits will be	0000...?
  // dummy-ign bits are:0111...?
  // disable ign:		1111....
  uint8_t ign_out;	// EDIS:0x0? dummy: 0x7? disable:0xff bit0: invertout

  // Fuelpump section:
  uint8_t engine_off_delay;		// turn off fuelpump after inactivity time [262msec]
  uint8_t pump_on_mintime;		// minimum time to leave fuelpump on at startup [262msec]
};

extern struct config_t config;
/***********************************************************************/

//config11
#define MAP_SENSOR      0  // 0:mpx4115ap, 1:mpx4250ap
#define ENGINE_STROKE   2  // 0:4-stroke, 1:2-stroke
#define INJ_TYPE        3  // 0:port injection, 1:throttle body
//bit 7-4 defines no of cylinders

/***********************************************************************/

//config12
//bit 1-0 coolant sensor type
//bit 3-2 mat sensor type
//bit 7-4 number of injectors

/***********************************************************************/

//config13
#define ODDFIRE               0 // 0:normal, 1:odd-fire
#define O2_WB_SENSOR          1 // 0:narrowband, 1:diy-wb
#define CONTROL_STRATEGY      2 // 0:speed-density, 1:alpha-N
#define BARO_CORRECTION       3 // 0:off, 1:on
//bit 7-4 unsused

/***********************************************************************/
// master_debug
#define mat2map        0    // reroute the mat sensor value to the map sensor value

/***********************************************************************/
// ve_conf
#define ve_learn_on    0   // enable the VE learning
#define ve_learn_sim   1   // only simulate VE learning
#define ve_learn_2x2   2   // modify all four elements

/***********************************************************************/
// lcd_c0
#define lcd_busypolled	0	// if 1, the busy flag is polled
#define lcd_easygoto	1 	// if 0, we force goto after each data
#define phasesignoff	2	// if 0 (and PHASESIGNAL is defined) you are dead
/***********************************************************************/
// trigger
#define trigger_edge   0       // if 1, then rising edge
#define ptrig_coil     1       // if 1, then use coil to trigger instead of toothwheel
#define strig_disable  1       // if 1, then secondary trigger is disabled
#define trigger_filter 2       // if 1, then filtering is enabled

/***********************************************************************/

// ign_out definitions:
#define IGN_DISABLE		(7)
#define IGN_DUMMYDWELL	(6)
#define IGN_BEFORE		(5)
#define IGN_REFSPARK	(4)
#define IGN_MULTISPARK	(3)
#define IGN_INVERTOUT	(0)
// used combinations of upper nibble (7..4) bits:
#define IGNTYPE_EDIS	(0x00)
#define IGNTYPE_DUMMY	(0x70)
// the rest 14 combinations are off, but:
#define IGNTYPE_NONE	(0xF0)


// These two shouln't be _here_
void setleds(void);
extern volatile uint8_t ledstate;       // interrupt sets this, not port directly. port is written from userspace 

// should be in <avr/pgmspace.h> but it's missing from older avrlibc:
#define PRG_RDW(a) ((PRG_RDB((unsigned char*)(a)) & 0x00FF) | ((PRG_RDB((unsigned char*)(a)+1))<<8))
#ifndef pgm_read_word
#define pgm_read_word(a) ((PRG_RDB((unsigned char*)(a)) & 0x00FF) | ((PRG_RDB((unsigned char*)(a)+1))<<8))
#define pgm_read_word_near(a) ((PRG_RDB((unsigned char*)(a)) & 0x00FF) | ((PRG_RDB((unsigned char*)(a)+1))<<8))
#define pgm_read_byte(a) (PRG_RDB((unsigned char*)(a)))
#endif	// pgm_read_word
extern const uint8_t BV_rev_arr[];

// similar to BV macro but reverse bitorder (0 => 0x80)
// extern uint8_t BV_rev_arr[] PROGMEM ; // = {0x80, 0x40, 0x20, 0x10, 8, 4, 2, 1};
#define BV_rev(a) (pgm_read_byte(&BV_rev_arr[a]))
#define BV_rev_l(a) (pgm_read_byte(&BV_rev_arr[a&7]))

#endif	// GLOBAL_H


